/****************************************************************************
** Meta object code from reading C++ file 'greeter.h'
**
** Created by: The Qt Meta Object Compiler version 67 (Qt 5.15.19)
**
** WARNING! All changes made in this file will be lost!
*****************************************************************************/

#include <memory>
#include "QLightDM/greeter.h"
#include <QtCore/qbytearray.h>
#include <QtCore/qmetatype.h>
#if !defined(Q_MOC_OUTPUT_REVISION)
#error "The header file 'greeter.h' doesn't include <QObject>."
#elif Q_MOC_OUTPUT_REVISION != 67
#error "This file was generated using the moc from 5.15.19. It"
#error "cannot be used with the include files from this version of Qt."
#error "(The moc has changed too much.)"
#endif

QT_BEGIN_MOC_NAMESPACE
QT_WARNING_PUSH
QT_WARNING_DISABLE_DEPRECATED
struct qt_meta_stringdata_QLightDM__Greeter_t {
    QByteArrayData data[50];
    char stringdata0[680];
};
#define QT_MOC_LITERAL(idx, ofs, len) \
    Q_STATIC_BYTE_ARRAY_DATA_HEADER_INITIALIZER_WITH_OFFSET(len, \
    qptrdiff(offsetof(qt_meta_stringdata_QLightDM__Greeter_t, stringdata0) + ofs \
        - idx * sizeof(QByteArrayData)) \
    )
static const qt_meta_stringdata_QLightDM__Greeter_t qt_meta_stringdata_QLightDM__Greeter = {
    {
QT_MOC_LITERAL(0, 0, 17), // "QLightDM::Greeter"
QT_MOC_LITERAL(1, 18, 11), // "showMessage"
QT_MOC_LITERAL(2, 30, 0), // ""
QT_MOC_LITERAL(3, 31, 4), // "text"
QT_MOC_LITERAL(4, 36, 30), // "QLightDM::Greeter::MessageType"
QT_MOC_LITERAL(5, 67, 4), // "type"
QT_MOC_LITERAL(6, 72, 10), // "showPrompt"
QT_MOC_LITERAL(7, 83, 29), // "QLightDM::Greeter::PromptType"
QT_MOC_LITERAL(8, 113, 22), // "authenticationComplete"
QT_MOC_LITERAL(9, 136, 21), // "autologinTimerExpired"
QT_MOC_LITERAL(10, 158, 4), // "idle"
QT_MOC_LITERAL(11, 163, 5), // "reset"
QT_MOC_LITERAL(12, 169, 19), // "connectToDaemonSync"
QT_MOC_LITERAL(13, 189, 11), // "connectSync"
QT_MOC_LITERAL(14, 201, 12), // "authenticate"
QT_MOC_LITERAL(15, 214, 8), // "username"
QT_MOC_LITERAL(16, 223, 19), // "authenticateAsGuest"
QT_MOC_LITERAL(17, 243, 21), // "authenticateAutologin"
QT_MOC_LITERAL(18, 265, 18), // "authenticateRemote"
QT_MOC_LITERAL(19, 284, 7), // "session"
QT_MOC_LITERAL(20, 292, 7), // "respond"
QT_MOC_LITERAL(21, 300, 8), // "response"
QT_MOC_LITERAL(22, 309, 20), // "cancelAuthentication"
QT_MOC_LITERAL(23, 330, 15), // "cancelAutologin"
QT_MOC_LITERAL(24, 346, 11), // "setLanguage"
QT_MOC_LITERAL(25, 358, 8), // "language"
QT_MOC_LITERAL(26, 367, 13), // "setResettable"
QT_MOC_LITERAL(27, 381, 10), // "resettable"
QT_MOC_LITERAL(28, 392, 16), // "startSessionSync"
QT_MOC_LITERAL(29, 409, 23), // "ensureSharedDataDirSync"
QT_MOC_LITERAL(30, 433, 13), // "authenticated"
QT_MOC_LITERAL(31, 447, 18), // "authenticationUser"
QT_MOC_LITERAL(32, 466, 14), // "defaultSession"
QT_MOC_LITERAL(33, 481, 10), // "selectUser"
QT_MOC_LITERAL(34, 492, 11), // "selectGuest"
QT_MOC_LITERAL(35, 504, 8), // "hostname"
QT_MOC_LITERAL(36, 513, 4), // "osId"
QT_MOC_LITERAL(37, 518, 6), // "osName"
QT_MOC_LITERAL(38, 525, 12), // "osPrettyName"
QT_MOC_LITERAL(39, 538, 9), // "osVersion"
QT_MOC_LITERAL(40, 548, 11), // "osVersionId"
QT_MOC_LITERAL(41, 560, 4), // "motd"
QT_MOC_LITERAL(42, 565, 15), // "hasGuestAccount"
QT_MOC_LITERAL(43, 581, 6), // "locked"
QT_MOC_LITERAL(44, 588, 10), // "PromptType"
QT_MOC_LITERAL(45, 599, 18), // "PromptTypeQuestion"
QT_MOC_LITERAL(46, 618, 16), // "PromptTypeSecret"
QT_MOC_LITERAL(47, 635, 11), // "MessageType"
QT_MOC_LITERAL(48, 647, 15), // "MessageTypeInfo"
QT_MOC_LITERAL(49, 663, 16) // "MessageTypeError"

    },
    "QLightDM::Greeter\0showMessage\0\0text\0"
    "QLightDM::Greeter::MessageType\0type\0"
    "showPrompt\0QLightDM::Greeter::PromptType\0"
    "authenticationComplete\0autologinTimerExpired\0"
    "idle\0reset\0connectToDaemonSync\0"
    "connectSync\0authenticate\0username\0"
    "authenticateAsGuest\0authenticateAutologin\0"
    "authenticateRemote\0session\0respond\0"
    "response\0cancelAuthentication\0"
    "cancelAutologin\0setLanguage\0language\0"
    "setResettable\0resettable\0startSessionSync\0"
    "ensureSharedDataDirSync\0authenticated\0"
    "authenticationUser\0defaultSession\0"
    "selectUser\0selectGuest\0hostname\0osId\0"
    "osName\0osPrettyName\0osVersion\0osVersionId\0"
    "motd\0hasGuestAccount\0locked\0PromptType\0"
    "PromptTypeQuestion\0PromptTypeSecret\0"
    "MessageType\0MessageTypeInfo\0"
    "MessageTypeError"
};
#undef QT_MOC_LITERAL

static const uint qt_meta_data_QLightDM__Greeter[] = {

 // content:
       8,       // revision
       0,       // classname
       0,    0, // classinfo
      23,   14, // methods
      14,  178, // properties
       2,  220, // enums/sets
       0,    0, // constructors
       0,       // flags
       6,       // signalCount

 // signals: name, argc, parameters, tag, flags
       1,    2,  129,    2, 0x06 /* Public */,
       6,    2,  134,    2, 0x06 /* Public */,
       8,    0,  139,    2, 0x06 /* Public */,
       9,    0,  140,    2, 0x06 /* Public */,
      10,    0,  141,    2, 0x06 /* Public */,
      11,    0,  142,    2, 0x06 /* Public */,

 // slots: name, argc, parameters, tag, flags
      12,    0,  143,    2, 0x0a /* Public */,
      13,    0,  144,    2, 0x0a /* Public */,
      14,    1,  145,    2, 0x0a /* Public */,
      14,    0,  148,    2, 0x2a /* Public | MethodCloned */,
      16,    0,  149,    2, 0x0a /* Public */,
      17,    0,  150,    2, 0x0a /* Public */,
      18,    2,  151,    2, 0x0a /* Public */,
      18,    1,  156,    2, 0x2a /* Public | MethodCloned */,
      18,    0,  159,    2, 0x2a /* Public | MethodCloned */,
      20,    1,  160,    2, 0x0a /* Public */,
      22,    0,  163,    2, 0x0a /* Public */,
      23,    0,  164,    2, 0x0a /* Public */,
      24,    1,  165,    2, 0x0a /* Public */,
      26,    1,  168,    2, 0x0a /* Public */,
      28,    1,  171,    2, 0x0a /* Public */,
      28,    0,  174,    2, 0x2a /* Public | MethodCloned */,
      29,    1,  175,    2, 0x0a /* Public */,

 // signals: parameters
    QMetaType::Void, QMetaType::QString, 0x80000000 | 4,    3,    5,
    QMetaType::Void, QMetaType::QString, 0x80000000 | 7,    3,    5,
    QMetaType::Void,
    QMetaType::Void,
    QMetaType::Void,
    QMetaType::Void,

 // slots: parameters
    QMetaType::Bool,
    QMetaType::Bool,
    QMetaType::Void, QMetaType::QString,   15,
    QMetaType::Void,
    QMetaType::Void,
    QMetaType::Void,
    QMetaType::Void, QMetaType::QString, QMetaType::QString,   19,   15,
    QMetaType::Void, QMetaType::QString,   19,
    QMetaType::Void,
    QMetaType::Void, QMetaType::QString,   21,
    QMetaType::Void,
    QMetaType::Void,
    QMetaType::Void, QMetaType::QString,   25,
    QMetaType::Void, QMetaType::Bool,   27,
    QMetaType::Bool, QMetaType::QString,   19,
    QMetaType::Bool,
    QMetaType::QString, QMetaType::QString,   15,

 // properties: name, type, flags
      30, QMetaType::Bool, 0x00095001,
      31, QMetaType::QString, 0x00095001,
      32, QMetaType::QString, 0x00095401,
      33, QMetaType::QString, 0x00095401,
      34, QMetaType::Bool, 0x00095401,
      35, QMetaType::QString, 0x00095401,
      36, QMetaType::QString, 0x00095401,
      37, QMetaType::QString, 0x00095401,
      38, QMetaType::QString, 0x00095401,
      39, QMetaType::QString, 0x00095401,
      40, QMetaType::QString, 0x00095401,
      41, QMetaType::QString, 0x00095401,
      42, QMetaType::Bool, 0x00095401,
      43, QMetaType::Bool, 0x00095401,

 // enums: name, alias, flags, count, data
      44,   44, 0x0,    2,  230,
      47,   47, 0x0,    2,  234,

 // enum data: key, value
      45, uint(QLightDM::Greeter::PromptTypeQuestion),
      46, uint(QLightDM::Greeter::PromptTypeSecret),
      48, uint(QLightDM::Greeter::MessageTypeInfo),
      49, uint(QLightDM::Greeter::MessageTypeError),

       0        // eod
};

void QLightDM::Greeter::qt_static_metacall(QObject *_o, QMetaObject::Call _c, int _id, void **_a)
{
    if (_c == QMetaObject::InvokeMetaMethod) {
        auto *_t = static_cast<Greeter *>(_o);
        (void)_t;
        switch (_id) {
        case 0: _t->showMessage((*reinterpret_cast< QString(*)>(_a[1])),(*reinterpret_cast< QLightDM::Greeter::MessageType(*)>(_a[2]))); break;
        case 1: _t->showPrompt((*reinterpret_cast< QString(*)>(_a[1])),(*reinterpret_cast< QLightDM::Greeter::PromptType(*)>(_a[2]))); break;
        case 2: _t->authenticationComplete(); break;
        case 3: _t->autologinTimerExpired(); break;
        case 4: _t->idle(); break;
        case 5: _t->reset(); break;
        case 6: { bool _r = _t->connectToDaemonSync();
            if (_a[0]) *reinterpret_cast< bool*>(_a[0]) = std::move(_r); }  break;
        case 7: { bool _r = _t->connectSync();
            if (_a[0]) *reinterpret_cast< bool*>(_a[0]) = std::move(_r); }  break;
        case 8: _t->authenticate((*reinterpret_cast< const QString(*)>(_a[1]))); break;
        case 9: _t->authenticate(); break;
        case 10: _t->authenticateAsGuest(); break;
        case 11: _t->authenticateAutologin(); break;
        case 12: _t->authenticateRemote((*reinterpret_cast< const QString(*)>(_a[1])),(*reinterpret_cast< const QString(*)>(_a[2]))); break;
        case 13: _t->authenticateRemote((*reinterpret_cast< const QString(*)>(_a[1]))); break;
        case 14: _t->authenticateRemote(); break;
        case 15: _t->respond((*reinterpret_cast< const QString(*)>(_a[1]))); break;
        case 16: _t->cancelAuthentication(); break;
        case 17: _t->cancelAutologin(); break;
        case 18: _t->setLanguage((*reinterpret_cast< const QString(*)>(_a[1]))); break;
        case 19: _t->setResettable((*reinterpret_cast< bool(*)>(_a[1]))); break;
        case 20: { bool _r = _t->startSessionSync((*reinterpret_cast< const QString(*)>(_a[1])));
            if (_a[0]) *reinterpret_cast< bool*>(_a[0]) = std::move(_r); }  break;
        case 21: { bool _r = _t->startSessionSync();
            if (_a[0]) *reinterpret_cast< bool*>(_a[0]) = std::move(_r); }  break;
        case 22: { QString _r = _t->ensureSharedDataDirSync((*reinterpret_cast< const QString(*)>(_a[1])));
            if (_a[0]) *reinterpret_cast< QString*>(_a[0]) = std::move(_r); }  break;
        default: ;
        }
    } else if (_c == QMetaObject::IndexOfMethod) {
        int *result = reinterpret_cast<int *>(_a[0]);
        {
            using _t = void (Greeter::*)(QString , QLightDM::Greeter::MessageType );
            if (*reinterpret_cast<_t *>(_a[1]) == static_cast<_t>(&Greeter::showMessage)) {
                *result = 0;
                return;
            }
        }
        {
            using _t = void (Greeter::*)(QString , QLightDM::Greeter::PromptType );
            if (*reinterpret_cast<_t *>(_a[1]) == static_cast<_t>(&Greeter::showPrompt)) {
                *result = 1;
                return;
            }
        }
        {
            using _t = void (Greeter::*)();
            if (*reinterpret_cast<_t *>(_a[1]) == static_cast<_t>(&Greeter::authenticationComplete)) {
                *result = 2;
                return;
            }
        }
        {
            using _t = void (Greeter::*)();
            if (*reinterpret_cast<_t *>(_a[1]) == static_cast<_t>(&Greeter::autologinTimerExpired)) {
                *result = 3;
                return;
            }
        }
        {
            using _t = void (Greeter::*)();
            if (*reinterpret_cast<_t *>(_a[1]) == static_cast<_t>(&Greeter::idle)) {
                *result = 4;
                return;
            }
        }
        {
            using _t = void (Greeter::*)();
            if (*reinterpret_cast<_t *>(_a[1]) == static_cast<_t>(&Greeter::reset)) {
                *result = 5;
                return;
            }
        }
    }
#ifndef QT_NO_PROPERTIES
    else if (_c == QMetaObject::ReadProperty) {
        auto *_t = static_cast<Greeter *>(_o);
        (void)_t;
        void *_v = _a[0];
        switch (_id) {
        case 0: *reinterpret_cast< bool*>(_v) = _t->isAuthenticated(); break;
        case 1: *reinterpret_cast< QString*>(_v) = _t->authenticationUser(); break;
        case 2: *reinterpret_cast< QString*>(_v) = _t->defaultSessionHint(); break;
        case 3: *reinterpret_cast< QString*>(_v) = _t->selectUserHint(); break;
        case 4: *reinterpret_cast< bool*>(_v) = _t->selectGuestHint(); break;
        case 5: *reinterpret_cast< QString*>(_v) = _t->hostname(); break;
        case 6: *reinterpret_cast< QString*>(_v) = _t->osId(); break;
        case 7: *reinterpret_cast< QString*>(_v) = _t->osName(); break;
        case 8: *reinterpret_cast< QString*>(_v) = _t->osPrettyName(); break;
        case 9: *reinterpret_cast< QString*>(_v) = _t->osVersion(); break;
        case 10: *reinterpret_cast< QString*>(_v) = _t->osVersionId(); break;
        case 11: *reinterpret_cast< QString*>(_v) = _t->motd(); break;
        case 12: *reinterpret_cast< bool*>(_v) = _t->hasGuestAccountHint(); break;
        case 13: *reinterpret_cast< bool*>(_v) = _t->lockHint(); break;
        default: break;
        }
    } else if (_c == QMetaObject::WriteProperty) {
    } else if (_c == QMetaObject::ResetProperty) {
    }
#endif // QT_NO_PROPERTIES
}

QT_INIT_METAOBJECT const QMetaObject QLightDM::Greeter::staticMetaObject = { {
    QMetaObject::SuperData::link<QObject::staticMetaObject>(),
    qt_meta_stringdata_QLightDM__Greeter.data,
    qt_meta_data_QLightDM__Greeter,
    qt_static_metacall,
    nullptr,
    nullptr
} };


const QMetaObject *QLightDM::Greeter::metaObject() const
{
    return QObject::d_ptr->metaObject ? QObject::d_ptr->dynamicMetaObject() : &staticMetaObject;
}

void *QLightDM::Greeter::qt_metacast(const char *_clname)
{
    if (!_clname) return nullptr;
    if (!strcmp(_clname, qt_meta_stringdata_QLightDM__Greeter.stringdata0))
        return static_cast<void*>(this);
    return QObject::qt_metacast(_clname);
}

int QLightDM::Greeter::qt_metacall(QMetaObject::Call _c, int _id, void **_a)
{
    _id = QObject::qt_metacall(_c, _id, _a);
    if (_id < 0)
        return _id;
    if (_c == QMetaObject::InvokeMetaMethod) {
        if (_id < 23)
            qt_static_metacall(this, _c, _id, _a);
        _id -= 23;
    } else if (_c == QMetaObject::RegisterMethodArgumentMetaType) {
        if (_id < 23)
            *reinterpret_cast<int*>(_a[0]) = -1;
        _id -= 23;
    }
#ifndef QT_NO_PROPERTIES
    else if (_c == QMetaObject::ReadProperty || _c == QMetaObject::WriteProperty
            || _c == QMetaObject::ResetProperty || _c == QMetaObject::RegisterPropertyMetaType) {
        qt_static_metacall(this, _c, _id, _a);
        _id -= 14;
    } else if (_c == QMetaObject::QueryPropertyDesignable) {
        _id -= 14;
    } else if (_c == QMetaObject::QueryPropertyScriptable) {
        _id -= 14;
    } else if (_c == QMetaObject::QueryPropertyStored) {
        _id -= 14;
    } else if (_c == QMetaObject::QueryPropertyEditable) {
        _id -= 14;
    } else if (_c == QMetaObject::QueryPropertyUser) {
        _id -= 14;
    }
#endif // QT_NO_PROPERTIES
    return _id;
}

// SIGNAL 0
void QLightDM::Greeter::showMessage(QString _t1, QLightDM::Greeter::MessageType _t2)
{
    void *_a[] = { nullptr, const_cast<void*>(reinterpret_cast<const void*>(std::addressof(_t1))), const_cast<void*>(reinterpret_cast<const void*>(std::addressof(_t2))) };
    QMetaObject::activate(this, &staticMetaObject, 0, _a);
}

// SIGNAL 1
void QLightDM::Greeter::showPrompt(QString _t1, QLightDM::Greeter::PromptType _t2)
{
    void *_a[] = { nullptr, const_cast<void*>(reinterpret_cast<const void*>(std::addressof(_t1))), const_cast<void*>(reinterpret_cast<const void*>(std::addressof(_t2))) };
    QMetaObject::activate(this, &staticMetaObject, 1, _a);
}

// SIGNAL 2
void QLightDM::Greeter::authenticationComplete()
{
    QMetaObject::activate(this, &staticMetaObject, 2, nullptr);
}

// SIGNAL 3
void QLightDM::Greeter::autologinTimerExpired()
{
    QMetaObject::activate(this, &staticMetaObject, 3, nullptr);
}

// SIGNAL 4
void QLightDM::Greeter::idle()
{
    QMetaObject::activate(this, &staticMetaObject, 4, nullptr);
}

// SIGNAL 5
void QLightDM::Greeter::reset()
{
    QMetaObject::activate(this, &staticMetaObject, 5, nullptr);
}
QT_WARNING_POP
QT_END_MOC_NAMESPACE
